from torchok.models.heads.classification.arcface_head import (
    ArcFaceHead,
)
from torchok.models.heads.classification.classification_head import (
    ClassificationHead,
)


__all__ = [
    'ArcFaceHead',
    'ClassificationHead',
]
